/* clockcheck.c - check clocks on my local net
 *
 * Copyright (c) 1998-2000 Tony Aiuto
 *
 * To report problems with this tool or to suggest enhancements send mail
 * to tony@aiu.to
 *
 */

#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 1998-2000, Tony Aiuto";
#endif

/*
 * Author: Tony Aiuto (tony@aiu.to)
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Edit History:
 * $Log: clockcheck.c $
 * Revision tools_TonyAiuto/2 2000/11/15 14:26:04 TonyAiuto
 * 	clean up and add rpm packaging
 * 
 * Revision 1.5  1999/04/23 05:06:05  tony
 * added -S option
 *
 * Revision 1.4  1998/06/24 13:47:57  tony
 * Add -q option, change display to fast/slow by nnn seconds
 *
 * Revision 1.3  1998/06/12 17:31:24  tony
 * do time delta on average of start & end host connect time
 *
 * Revision 1.2  1998/06/12 17:28:25  tony
 * convert to use time (port 37) service & add clock differences
 *
 * Revision 1.1  1998/02/11 23:20:04  tony
 * Initial revision
 *
 * Revision 1.1  1998/02/11 22:06:13  tony
 * Initial revision
 *
 * Revision 1.2  1997/12/30 19:43:45  tony
 * Add -r option
 *
 * Revision 1.1  1997/12/24 17:07:11  tony
 * Initial revision
 *
 *
 * 12/22/97 tony - created
 */

#ifndef lint
static char rcsid[] = "@(#) $Id: clockcheck.c tools_TonyAiuto/2 2000/11/15 14:26:04 TonyAiuto $";
#endif


#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#define _BSD_SIGNALS 1

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define 	DISPLAY_SIGNED_TIME 0
#define		DEBUG 0


int ConnectHost(const char *);
time_t GetTime(const char *host);


int		verbose = 1;
int		debug = 0;
#if USE_DAYTIME
static	int	timePort = 13;
#else
static	int	timePort = 37;
#endif
static	int	daytime_socket = -1;
static	time_t	beginTime;		/* time we started */
static	time_t	baseTime;		/* time we started */
static	long	threshold = 0;		/* allowed offset */

#define	TIME_BASE_OFFSET 2208988800ul;	/* 1/1/1900 to 1/1/1970 */


extern	int	optind;
extern	char	*optarg;

static void usage()
{
#if !defined(VERSION)
#define	VERSION	"unknown"
#endif
	printf("clockcheck %s\n", VERSION);
	printf("usage: checkclock [-q] [-t threshold] [-p port] [-S] host ...\n");
	printf("\t-q\tQuiet mode\n");
	printf("\t-S\tSet clock to time from remote sites.\n");
	exit(0);
}


void shutmedown(int sig)
{
	printf("Requested shutdown\n");
	if(daytime_socket >= 0) close(daytime_socket);
	exit(0);
}

char *ptime(time_t *aTime)
{
	static char	obuf[100];
	struct tm *tmbuf;
	tmbuf = localtime(aTime);
#if USE_DAYTIME
	strftime(obuf, sizeof(obuf), "%a %b %d %H:%M:%S %Y", tmbuf);
#else
	strftime(obuf, sizeof(obuf), "%Y-%m-%d %H:%M:%S", tmbuf);
#endif

	return obuf;
}

#
static void setup()
{
#if THIS_CODE_IS_NOT_WORKING_BUT_I_DO_NOT_KNOW_WHY
	struct servent *timeservice = NULL;
	setservent(0);
	timeservice = getservbyname("time", "tcp");
	if(timeservice != NULL) {
		timePort = timeservice->s_port;
		printf("%s port is %d\n", timeservice->s_name, timePort);
	}
	endservent();
#endif
}

main(int argc, char *argv[])
{
	int	c;
	char	myhost[50];
	time_t	htimes;		// average difference from each reference host
	int	nHosts;
	int	setClock = 0;

	setup();

	while((c = getopt(argc, argv, "dp:qt:vS")) != EOF) {
		switch(c) {
		case 'd':
			debug++;
			break;
		case 'p':
			timePort = atoi(optarg);
			break;

		case 'q':
			verbose = 0;
			break;

		case 'S':
			setClock = 1;
			break;

		case 't':
			threshold = atol(optarg);
			break;

		case 'v':
			verbose++;
			break;


		default:
			usage();
		}
	}
	if(argc - optind < 1) {
		usage();
	}

	signal(SIGTERM, shutmedown);
	signal(SIGQUIT, shutmedown);
	signal(SIGINT, shutmedown);

#if USE_DAYTIME
	printf("Just printing times, no time difference checks.\n");
#endif
	time(&beginTime);
	baseTime = beginTime;
	gethostname(myhost, sizeof(myhost));
#if DEBUG
	printf("localtime is %ld\n", beginTime);
#endif
	printf("%-20.20s: %s\n", myhost, ptime(&beginTime));

	htimes = 0;
	nHosts = 0;
	for( ; optind < argc; optind++) {
		daytime_socket = ConnectHost(argv[optind]);
		/* Track time we started asking the host */
		time(&beginTime);
		if(daytime_socket >= 0) {
			htimes += GetTime(argv[optind]);
			close(daytime_socket);
			daytime_socket = -1;
			nHosts += 1;
		}
	}
	htimes /= nHosts;

	if(setClock) {
  		struct timeval ct;
		gettimeofday(&ct, NULL);
		printf("adjusting local time by %d seconds\n", htimes);
		ct.tv_sec -= htimes;
		settimeofday(&ct, NULL);
	}

	return 0;
}


int ConnectHost(const char *host)
{
	int	sock;
	struct sockaddr_in sin;

	if (strlen(host) > 255) {
		fprintf(stderr,
			"Requested daytime hostname too long: %.200s.",
			host);
		return -1;
	}
  
	memset((char *) &sin, 0, sizeof(sin));
#ifdef BROKEN_INET_ADDR
	sin.sin_addr.s_addr = inet_network(host);
#else /* BROKEN_INET_ADDR */
	sin.sin_addr.s_addr = inet_addr(host);
#endif /* BROKEN_INET_ADDR */
	if ((sin.sin_addr.s_addr & 0xffffffff) != 0xffffffff) {
		/* It was a valid numeric host address. */
		sin.sin_family = AF_INET;
	} else {
		/* Look up the host address from the name servers. */
		struct hostent *hp = gethostbyname(host);
		if (!hp) {
			printf("%s: unknown host.\n", host);
	  		return -1;
		}
		if (!hp->h_addr_list[0]) {
			printf("%s: host has no IP address.\n", host);
			return -1;
		}
		sin.sin_family = hp->h_addrtype;
		memcpy((char *) &sin.sin_addr, hp->h_addr_list[0], sizeof(sin.sin_addr));
	}
	sin.sin_port = htons(timePort);

	/* Create the socket. */
	sock = socket(sin.sin_family, SOCK_STREAM, 0);
	if (sock < 0) {
		printf("socket: %s\n", strerror(errno));
		return -2;
	}
	if(debug > 1) printf("Got socket\n");


	/* Connect to the host/port. */
	if (connect(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		printf("connect %s:%d: %s\n", host, timePort, strerror(errno));
		close(sock);
		return -1;
	}
	if(debug > 0) printf("Connected\n");

	/* Successful connection. */

#if TODO
	// We minght want to set non blocking if we want it
	// to be interruptable

#if defined(O_NONBLOCK) && !defined(O_NONBLOCK_BROKEN)
  (void)fcntl(sock, F_SETFL, O_NONBLOCK);
#else /* O_NONBLOCK && !O_NONBLOCK_BROKEN */
  (void)fcntl(sock, F_SETFL, O_NDELAY);
#endif /* O_NONBLOCK && !O_NONBLOCK_BROKEN */

#endif

	return sock;
}



/** GetTime actually does a few things.
 *	It gets and prints the remote time
 *	It returns the clock difference from local to remote
 */

time_t GetTime(const char *host)
{
	int	n, j;
	time_t	localTime;
	time_t	remoteTime;
	time_t	delta;
	unsigned char	resp[50];

#if NO
	strcpy(cmd, "\r\n");
	if(debug) printf("Sending '%s'\n", cmd);
	write(daytime_socket, cmd, strlen(cmd));
#endif

	/* Get the response */
	while(1) {
		long tripDiff;
		n = read(daytime_socket, resp, sizeof(resp)-1);

		if(n < 0) break;
		if(n == 0) continue;
		resp[n] = '\0';

#if USE_DAYTIME
		printf("%-20.20s: ", host);
		fwrite(resp, 1, n, stdout);
#else
		time(&localTime);

		/* average local begin and end as an estimate of what
		 * the remote time SHOULD be
		 */
#if DEBUG
		printf("begin=%ld, remote=%d", beginTime, localTime);
#endif
		// This is the obvious way, but if starts to overflow
		// around Jan 11, 2004 (as I found out)

		// localTime = (beginTime + localTime)/2;

		tripDiff = beginTime - localTime;
		localTime = beginTime + tripDiff / 2;
#if DEBUG
		printf(", avg=%ld\n", localTime);
#endif
		remoteTime = 0;
		for(j = 0; j < n; j++) {
			/* printf(" %x", resp[j]); */
			remoteTime = (remoteTime << 8) | (resp[j] & 0xff);
		}
		remoteTime  -= TIME_BASE_OFFSET;
		/* printf(" Remote time is %x=%ld", remoteTime, remoteTime); */
		delta = localTime - remoteTime;
#if DISPLAY_SIGNED_TIME
		if(delta > threshold || delta < -threshold) {
			printf("%-20.20s: %s", host, ptime(&remoteTime));
			printf(" off by %d seconds", delta);
		} else {
			if(verbose > 0) {
			    printf("%-20.20s: %s", host, ptime(&remoteTime));
			}
		}
		printf("\n");
#else
		if(delta > threshold) {
			printf("%-20.20s: %s", host, ptime(&remoteTime));
			printf(" slow by %d seconds\n", delta);
		} else if(delta < -threshold) {
			printf("%-20.20s: %s", host, ptime(&remoteTime));
			printf(" fast by %d seconds\n", -delta);
		} else {
			if(verbose > 0) {
			    printf("%-20.20s: %s\n", host, ptime(&remoteTime));
			}
		}
#endif /* DISPLAY_SIGNED_TIME */
#endif
		break;
	}
	return delta;
}
